from collections import defaultdict
from enum import Enum
import functools
from typing import (
    Callable,
    Dict,
    Generic,
    Iterable,
    Mapping,
    Optional,
    Set,
    TypeVar,
    Union,
)

import pydantic
from trulens.core.utils import python as python_utils
from trulens.core.utils import text as text_utils

T = TypeVar("T")


class Feature(str, Enum):
    """Experimental feature flags.

    Use [TruSession.experimental_enable_feature][trulens.core.session.TruSession.experimental_enable_feature] to enable
    these features:

    Examples:
        ```python
        from trulens.core.session import TruSession
        from trulens.core.experimental import Feature

        session = TruSession()

        session.experimental_enable_feature(Feature.OTEL_TRACING)
        ```
    """

    OTEL_TRACING = "otel_tracing"
    """OTEL-like tracing.

    !!! Warning
        This changes how wrapped functions are processed. This setting cannot be
        changed after any wrapper is produced.
    """

    @classmethod
    def _repr_all(cls) -> str:
        """Return a string representation of all the feature flags."""

        ret = ""
        for flag in cls:
            ret += f'  {flag.name} = "{flag.value}"\n'

        return ret

    @classmethod
    def _missing_(cls, value: str):
        raise ValueError(
            f"Invalid feature flag `{value}`. Available flags are:\n{cls._repr_all()}"
        )


class _Setting(Generic[T]):
    """A setting that attains some value and can be locked from changing."""

    def __init__(self, default: T):
        self.value: T = default
        """The stored value."""

        self.locked_by: Set[str] = set()
        """Set of representations of frames (not in trulens) that have locked this value.

        If empty, it has not been locked and can be changed.
        """

    @property
    def is_locked(self) -> bool:
        """Determine if the setting is locked."""

        return len(self.locked_by) > 0

    def set(self, value: Optional[T] = None, lock: bool = False) -> T:
        """Set/Get the value.

        Set the value first if a value is provided. Lock if lock is set.

        Raises:
            ValueError: If the setting has already been locked and the value is
                different from the current value.
        """

        if value is not None:
            if len(self.locked_by) > 0 and value != self.value:
                locked_frames = ""
                for frame_str in self.locked_by:
                    locked_frames += f"  {frame_str}\n"

                raise ValueError(
                    f"Feature flag has already been set to {self.value} and cannot be changed. "
                    f"It has been locked here:\n{locked_frames}"
                )

            self.value = value

        if lock:
            frame = python_utils.external_caller_frame(offset=1)
            frame_str = python_utils.code_line(frame, show_source=True)
            # Store representation instead of frame to avoid keeping references
            # that would prevent GC.
            self.locked_by.add(frame_str)

        return self.value

    def get(self, lock: bool = False) -> T:
        """Get the value of this setting.

        If lock is True, lock the setting.
        """

        return self.set(lock=lock)

    def lock(self, value: Optional[T] = None) -> T:
        """Lock the value of this setting.

        If a value is provided, attempt to set the setting first to that value.

        Raises:
            ValueError: If the setting has already been locked and the value is
                different from the current value.
        """

        return self.set(value, lock=True)


class _Settings:
    """A collection of settings to enable/disable experimental features.

    A feature can be enabled/disabled and/or locked so that once it is set, it
    cannot be changed again. Locking is necessary for some features like OTEL
    tracing as once components have been instrumented with old or new tracing,
    the instrumentation cannot be changed.
    """

    def __init__(self):
        self.settings: Dict[Feature, _Setting[bool]] = defaultdict(
            lambda: _Setting(default=False)
        )
        """The settings for the experimental features."""

    def is_locked(self, flag: Union[Feature, str]) -> bool:
        """Determine if the given setting is locked."""

        return self.settings[Feature(flag)].is_locked

    def set(
        self,
        flag: Union[Feature, str],
        *,
        value: Optional[bool] = None,
        lock: bool = False,
    ) -> bool:
        """Get/Set the given feature flag to the given value.

        Sets the flag to the given value if the value parameter is set. Locks
        the flag if the lock parameter is set to True.

        Raises:
            ValueError: If the flag was already locked to a different value.
        """

        return self.settings[Feature(flag)].set(value=value, lock=lock)

    def lock(
        self, flag: Union[Feature, str], *, value: Optional[bool] = None
    ) -> bool:
        """Lock the given feature flag to the given value.

        If the value is not provided, lock the flag to its current value.

        Raises:
            ValueError: If the flag has already been locked to a different value.
        """

        return self.set(flag, value=value, lock=True)

    def get(self, flag: Union[str, Feature], *, lock: bool = False) -> bool:
        """Determine the value of the given feature flag by checking both global
        and instance flags.

        Instance value takes precedence over the global value.
        """

        return self.set(flag, lock=lock)

    def enable(self, flag: Union[Feature, str], *, lock: bool = False) -> bool:
        """Enable the given feature flag.

        Locks the flag if the lock parameter is set to True.

        Raises:
            ValueError: If the flag was already locked to disabled.
        """

        return self.set(flag, value=True, lock=lock)

    def disable(self, flag: Union[Feature, str], *, lock: bool = False) -> bool:
        """Disable the given feature flag.

        Locks the flag if the lock parameter is set to True.

        Raises:
            ValueError: If the flag was already locked to enabled.
        """

        return self.set(flag, value=False, lock=lock)

    def set_multiple(
        self,
        flags: Union[
            Iterable[Union[str, Feature]],
            Mapping[Union[str, Feature], bool],
        ],
        lock: bool = False,
    ):
        """Set multiple feature flags.

        If lock is set, lock the flags. If a dictionary is passed, the keys are
        the feature flags and the values are the values to set them to. If a
        list is passed, the flags are set to True.

        Raises:
            ValueError: If any of the flags are already locked to a different
                value than specified.
        """

        if isinstance(flags, dict):
            for flag, val in flags.items():
                self.set(flag, value=val, lock=lock)
        else:
            for flag in flags:
                self.set(flag, value=True, lock=lock)


class _WithExperimentalSettings(
    pydantic.BaseModel,
    text_utils.WithIdentString,
):
    """Mixin to add experimental flags and control methods.

    Prints out messages when features are enabled/disabled locked and when
    a setting fails to take up due to locking.
    """

    _experimental_feature_flags: _Settings = pydantic.PrivateAttr(
        default_factory=_Settings
    )
    """EXPERIMENTAL: Flags to control experimental features."""

    def _experimental_feature(
        self,
        flag: Union[str, Feature],
        *,
        value: Optional[bool] = None,
        lock: bool = False,
    ) -> bool:
        """Get and/or set the value of the given feature flag.

        Set it first if value is given. Lock it if lock is set.

        Raises:
            ValueError: If the flag is locked to a different value.
        """

        # NOTE(piotrm): The printouts are important as we want to make sure the
        # user is aware that they are using a experimental feature.

        flag = Feature(flag)

        was_locked = self._experimental_feature_flags.is_locked(flag)

        original_value = self._experimental_feature_flags.get(flag)
        val = self._experimental_feature_flags.set(flag, value=value, lock=lock)
        changed = val != original_value

        if value is not None and changed:
            if val:
                print(
                    f"{text_utils.UNICODE_CHECK} experimental {flag} enabled for {self._ident_str()}."
                )
            else:
                print(
                    f"{text_utils.UNICODE_STOP} experimental {flag} disabled for {self._ident_str()}"
                )

        if val and lock and not was_locked:
            print(
                f"{text_utils.UNICODE_LOCK} experimental {flag} is enabled and cannot be changed."
            )

        return val

    def _experimental_lock_feature(self, flag: Union[str, Feature]) -> bool:
        """Get and lock the given feature flag."""

        return self._experimental_feature(flag, lock=True)

    def experimental_enable_feature(self, flag: Union[str, Feature]) -> bool:
        """Enable the given feature flag.

        Raises:
            ValueError: If the flag is already locked to disabled.
        """

        return self._experimental_feature(flag, value=True)

    def experimental_disable_feature(self, flag: Union[str, Feature]) -> bool:
        """Disable the given feature flag.

        Raises:
            ValueError: If the flag is already locked to enabled.
        """

        return self._experimental_feature(flag, value=False)

    def experimental_feature(
        self, flag: Union[str, Feature], *, lock: bool = False
    ) -> bool:
        """Determine the value of the given feature flag.

        If lock is set, the flag will be locked to the value returned.
        """

        return self._experimental_feature(flag, lock=lock)

    def experimental_set_features(
        self,
        flags: Union[
            Iterable[Union[str, Feature]],
            Mapping[Union[str, Feature], bool],
        ],
        lock: bool = False,
    ):
        """Set multiple feature flags.

        If lock is set, the flags will be locked to the values given.

        Raises:
            ValueError: If any flag is already locked to a different value than
            provided.
        """

        if isinstance(flags, dict):
            for flag, val in flags.items():
                self._experimental_feature(flag, value=val, lock=lock)
        else:
            for flag in flags:
                self._experimental_feature(flag, value=True, lock=lock)

    def _experimental_assert_feature(
        self, flag: Feature, purpose: Optional[str] = None
    ):
        """Raise a ValueError if the given feature flag is not enabled.

        Gives instructions on how to enable the feature flag if error gets
        raised."""

        flag = Feature(flag)

        if purpose is None:
            purpose = "."
        else:
            purpose = f" for {purpose}."

        if not self.experimental_feature(flag):
            raise ValueError(
                f"""Feature flag {flag} is not enabled{purpose} You can enable it in two ways:
    ```python
    from trulens.core.experimental import Feature

    # Enable for this instance when creating it:
    val = {self.__class__.__name__}(experimental_feature_flags=[{flag}]

    # Enable for this instance after it has been created (for features that allows this):
    val.experimental_enable_feature({flag})
    ```
"""
            )

    @staticmethod
    def _experimental_method_override(
        flag: Feature, enabled: T, lock: bool = False
    ) -> T:
        """Decorator to replace the decorated method with the given one if the
        specified feature is enabled.

        Locks the flag if the lock parameter is set to True.

        Example:
            ```python
            class MyClass(WithExperimentalSettings, ...):

                def my_method_experimental(self, ...): ...

                @MyClass.experimental_method_override(
                    flag=Feature.OTEL_TRACING,
                    enabled=my_method_experimental
                )
                def my_method(self, ...): ...
            ```
        """

        def wrapper(func: T) -> T:
            @functools.wraps(func)
            def wrapped(self: _WithExperimentalSettings, *args, **kwargs):
                if self.experimental_feature(flag, lock=lock):
                    return enabled(*args, **kwargs)

                return func(self, *args, **kwargs)

            return wrapped

        return wrapper

    @staticmethod
    def _experimental_method(
        flag: Feature, enabled: Callable, disabled: Callable, lock: bool = False
    ) -> Callable:
        """Select between two methods based on the status of a feature flag.

        The selection happens after the method is called. Locks the flag if the lock
        parameter is set to True.

        Example:
            ```python
            class MyClass(WithExperimentalSettings, ...):
                ...
                def my_method_default(self, ...): ...
                def my_method_experimental(self, ...): ...
                ...
                my_method = MyClass.experimental_method(
                    flag=Feature.OTEL_TRACING,
                    enabled=my_method_experimental,
                    disabled=my_method_default
                )
                ```
        """

        @functools.wraps(enabled)  # or disabled
        def wrapper(self: _WithExperimentalSettings, *args, **kwargs):
            if self.experimental_feature(flag, lock=lock):
                return enabled(*args, **kwargs)

            return disabled(*args, **kwargs)

        return wrapper



"""OTEL Compatibility Classes

This module contains classes to support interacting with the OTEL ecosystem.
Additions on top of these meant for TruLens uses outside of OTEL compatibility
are found in `traces.py`.
"""

from __future__ import annotations

import contextlib
import contextvars
import logging
import random
import time
from types import TracebackType
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from opentelemetry import context as context_api
from opentelemetry import trace as trace_api
from opentelemetry.sdk import resources as resources_sdk
from opentelemetry.sdk import trace as trace_sdk
from opentelemetry.util import types as types_api
import pydantic
from trulens.core._utils.pycompat import NoneType  # import style exception
from trulens.core._utils.pycompat import TypeAlias  # import style exception
from trulens.core._utils.pycompat import TypeAliasType  # import style exception
from trulens.core.utils import python as python_utils
from trulens.core.utils import serial as serial_utils

logger = logging.getLogger(__name__)

# Type alises

A = TypeVar("A")
B = TypeVar("B")

TSpanID: TypeAlias = int
"""Type of span identifiers.
64 bit int as per OpenTelemetry.
"""
NUM_SPANID_BITS: int = 64
"""Number of bits in a span identifier."""

TTraceID: TypeAlias = int
"""Type of trace identifiers.
128 bit int as per OpenTelemetry.
"""
NUM_TRACEID_BITS: int = 128
"""Number of bits in a trace identifier."""

TTimestamp: TypeAlias = int
"""Type of timestamps in spans.

64 bit int representing nanoseconds since epoch as per OpenTelemetry.
"""
NUM_TIMESTAMP_BITS = 64

TLensedBaseType: TypeAlias = Union[str, int, float, bool]
"""Type of base types in span attributes.

!!! Warning
    OpenTelemetry does not allow None as an attribute value. Handling None is to
    be decided.
"""

TLensedAttributeValue = TypeAliasType(
    "TLensedAttributeValue",
    Union[
        str,
        int,
        float,
        bool,
        NoneType,  # TODO(SNOW-1711929): None is not technically allowed as an attribute value.
        Sequence["TLensedAttributeValue"],  # type: ignore
        "TLensedAttributes",
    ],
)
"""Type of values in span attributes."""

# NOTE(piotrm): pydantic will fail if you specify a recursive type alias without
# the TypeAliasType schema as above.

TLensedAttributes: TypeAlias = Dict[str, TLensedAttributeValue]
"""Attribute dictionaries.

Note that this deviates from what OTEL allows as attribute values. Because OTEL
does not allow general recursive values to be stored as attributes, we employ a
system of flattening values before exporting to OTEL. In this process we encode
a single generic value as multiple attributes where the attribute name include
paths/lenses to the parts of the generic value they are representing. For
example, an attribute/value like `{"a": {"b": 1, "c": 2}}` would be encoded as
`{"a.b": 1, "a.c": 2}`. This process is implemented in the
`flatten_lensed_attributes` method.
"""


def flatten_value(
    v: TLensedAttributeValue, lens: Optional[serial_utils.Lens] = None
) -> Iterable[Tuple[serial_utils.Lens, types_api.AttributeValue]]:
    """Flatten recursive value into OTEL-compatible attribute values.

    See `TLensedAttributes` for more details.
    """

    if lens is None:
        lens = serial_utils.Lens()

    # TODO(SNOW-1711929): OpenTelemetry does not allow None as an attribute
    # value. Unsure what is best to do here.

    # if v is None:
    #    yield (path, "None")

    elif v is None:
        pass

    elif isinstance(v, TLensedBaseType):
        yield (lens, v)

    elif isinstance(v, Sequence) and all(
        isinstance(e, TLensedBaseType) for e in v
    ):
        yield (lens, v)

    elif isinstance(v, Sequence):
        for i, e in enumerate(v):
            yield from flatten_value(v=e, lens=lens[i])

    elif isinstance(v, Mapping):
        for k, e in v.items():
            yield from flatten_value(v=e, lens=lens[k])

    else:
        raise ValueError(
            f"Do not know how to flatten value of type {type(v)} to OTEL attributes."
        )


def flatten_lensed_attributes(
    m: TLensedAttributes,
    path: Optional[serial_utils.Lens] = None,
    prefix: str = "",
) -> types_api.Attributes:
    """Flatten lensed attributes into OpenTelemetry attributes."""

    if path is None:
        path = serial_utils.Lens()

    ret = {}
    for k, v in m.items():
        if k.startswith(prefix):
            # Only flattening those attributes that begin with `prefix` are
            # those are the ones coming from trulens_eval.
            for p, a in flatten_value(v, path[k]):
                ret[str(p)] = a
        else:
            ret[k] = v

    return ret


def new_trace_id():
    return int(
        random.getrandbits(NUM_TRACEID_BITS)
        & trace_api.span._TRACE_ID_MAX_VALUE
    )


def new_span_id():
    return int(
        random.getrandbits(NUM_SPANID_BITS) & trace_api.span._SPAN_ID_MAX_VALUE
    )


class TraceState(serial_utils.SerialModel, trace_api.span.TraceState):
    """[OTEL TraceState][opentelemetry.trace.TraceState] requirements."""

    # Hackish: trace_api.span.TraceState uses _dict internally.
    _dict: Dict[str, str] = pydantic.PrivateAttr(default_factory=dict)


class SpanContext(serial_utils.SerialModel):
    """[OTEL SpanContext][opentelemetry.trace.SpanContext] requirements."""

    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    trace_id: int = pydantic.Field(default_factory=new_trace_id)
    """Unique identifier for the trace.

    Each root span has a unique trace id."""

    span_id: int = pydantic.Field(default_factory=new_span_id)
    """Identifier for the span.

    Meant to be at least unique within the same trace_id.
    """

    trace_flags: trace_api.TraceFlags = pydantic.Field(
        trace_api.DEFAULT_TRACE_OPTIONS
    )

    trace_state: TraceState = pydantic.Field(default_factory=TraceState)

    is_remote: bool = False

    _tracer: Tracer = pydantic.PrivateAttr(None)

    @property
    def tracer(self) -> Tracer:
        return self._tracer

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        for k, v in kwargs.items():
            if v is None:
                continue
            # pydantic does not set private attributes in init
            if k.startswith("_") and hasattr(self, k):
                setattr(self, k, v)


def lens_of_flat_key(key: str) -> serial_utils.Lens:
    """Convert a flat dict key to a lens."""
    lens = serial_utils.Lens()
    for step in key.split("."):
        lens = lens[step]

    return lens


class Span(serial_utils.SerialModel, trace_api.Span):
    """[OTEL Span][opentelemetry.trace.Span] requirements.

    See also [OpenTelemetry
    Span](https://opentelemetry.io/docs/specs/otel/trace/api/#span) and
    [OpenTelemetry Span
    specification](https://github.com/open-telemetry/opentelemetry-specification/blob/main/specification/trace/api.md).
    """

    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    _name: str = pydantic.PrivateAttr(None)

    @property
    def name(self) -> str:
        """Name of the span."""

        return self._name

    _kind: trace_api.SpanKind = pydantic.PrivateAttr(
        trace_api.SpanKind.INTERNAL
    )

    @property
    def kind(self) -> trace_api.SpanKind:
        """Kind of span."""

        return self._kind

    _context: SpanContext = pydantic.PrivateAttr(None)

    @property
    def context(self) -> SpanContext:
        """Trace/span identifiers."""

        return self._context

    _parent: Optional[SpanContext] = pydantic.PrivateAttr(None)

    @property
    def parent(self) -> Optional[SpanContext]:
        """Optional parent identifiers."""

        return self._parent

    _status: trace_api.status.StatusCode = pydantic.PrivateAttr(
        default=trace_api.status.StatusCode.UNSET
    )

    @property
    def status(self) -> trace_api.status.StatusCode:
        """Status of the span as per OpenTelemetry Span requirements."""

        return self._status

    _status_description: Optional[str] = pydantic.PrivateAttr(None)

    @property
    def status_description(self) -> Optional[str]:
        """Status description as per OpenTelemetry Span requirements."""

        return self._status_description

    _events: List[Tuple[str, trace_api.types.Attributes, TTimestamp]] = (
        pydantic.PrivateAttr(default_factory=list)
    )

    @property
    def events(
        self,
    ) -> List[Tuple[str, trace_api.types.Attributes, TTimestamp]]:
        """Events recorded in the span.

        !!! Warning

            Events in OpenTelemetry seem to be synonymous to logs. Do not store
            anything we want to query or process in events.
        """
        return self._events

    _links: Dict[SpanContext, trace_api.types.Attributes] = (
        pydantic.PrivateAttr(default_factory=dict)
    )

    @property
    def links(self) -> trace_api._Links:
        """Relationships to other spans with attributes on each link."""

        return self._links

    _attributes: trace_api.types.Attributes = pydantic.PrivateAttr(
        default_factory=dict
    )

    @property
    def attributes(self) -> trace_api.types.Attributes:
        """Data fields."""
        return self._attributes

    _start_timestamp: int = pydantic.PrivateAttr(default_factory=time.time_ns)

    @property
    def start_timestamp(self) -> int:
        """Start time in nanoseconds since epoch."""

        return self._start_timestamp

    _end_timestamp: Optional[int] = pydantic.PrivateAttr(None)

    @property
    def end_timestamp(self) -> Optional[int]:
        """End time in nanoseconds since epoch. None if not yet finished."""

        return self._end_timestamp

    _record_exception: bool = pydantic.PrivateAttr(True)
    """Whether to record exceptions in the span."""

    _set_status_on_exception: bool = pydantic.PrivateAttr(True)
    """Whether to set status to ERROR on exception."""

    _tracer: Tracer = pydantic.PrivateAttr(None)
    """NON-STANDARD: The Tracer that produced this span."""

    @property
    def tracer(self) -> Tracer:
        return self._tracer

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        for k, v in kwargs.items():
            if v is None:
                continue
            # pydantic does not set private attributes in init
            if k.startswith("_") and hasattr(self, k):
                setattr(self, k, v)

        if self._start_timestamp is None:
            self._start_timestamp = time.time_ns()

    def update_name(self, name: str) -> None:
        """See [OTEL update_name][opentelemetry.trace.span.Span.update_name]."""

        self._name = name

    def get_span_context(self) -> trace_api.span.SpanContext:
        """See [OTEL get_span_context][opentelemetry.trace.span.Span.get_span_context]."""

        return self.context

    def set_status(
        self,
        status: Union[trace_api.span.Status, trace_api.span.StatusCode],
        description: Optional[str] = None,
    ) -> None:
        """See [OTEL set_status][opentelemetry.trace.span.Span.set_status]."""

        if isinstance(status, trace_api.span.Status):
            if description is not None:
                raise ValueError(
                    "Ambiguous status description provided both in `status.description` and in `description`."
                )

            self._status = status.status_code
            self._status_description = status.description
        else:
            self._status = status
            self._status_description = description

    def add_event(
        self,
        name: str,
        attributes: types_api.Attributes = None,
        timestamp: Optional[TTimestamp] = None,
    ) -> None:
        """See [OTEL add_event][opentelemetry.trace.span.Span.add_event]."""

        self._events.append((name, attributes, timestamp or time.time_ns()))

    def add_link(
        self,
        context: trace_api.span.SpanContext,
        attributes: types_api.Attributes = None,
    ) -> None:
        """See [OTEL add_link][opentelemetry.trace.span.Span.add_link]."""

        if attributes is None:
            attributes = {}

        self._links[context] = attributes

    def is_recording(self) -> bool:
        """See [OTEL is_recording][opentelemetry.trace.span.Span.is_recording]."""

        return self._status == trace_api.status.StatusCode.UNSET

    def set_attributes(
        self, attributes: Dict[str, types_api.AttributeValue]
    ) -> None:
        """See [OTEL set_attributes][opentelemetry.trace.span.Span.set_attributes]."""

        for key, value in attributes.items():
            self.set_attribute(key, value)

    def set_attribute(self, key: str, value: types_api.AttributeValue) -> None:
        """See [OTEL set_attribute][opentelemetry.trace.span.Span.set_attribute]."""

        self.attributes[key] = value

    def record_exception(
        self,
        exception: BaseException,
        attributes: types_api.Attributes = None,
        timestamp: Optional[TTimestamp] = None,
        escaped: bool = False,  # purpose unknown
    ) -> None:
        """See [OTEL record_exception][opentelemetry.trace.span.Span.record_exception]."""

        if self._set_status_on_exception:
            self.set_status(
                trace_api.status.Status(trace_api.status.StatusCode.ERROR)
            )

        if self._record_exception:
            if attributes is None:
                attributes = {}

            attributes["exc_type"] = python_utils.class_name(type(exception))
            attributes["exc_val"] = str(exception)
            if exception.__traceback__ is not None:
                attributes["code_line"] = python_utils.code_line(
                    exception.__traceback__.tb_frame, show_source=True
                )

            self.add_event("trulens.exception", attributes, timestamp)

    def end(self, end_time: Optional[TTimestamp] = None):
        """See [OTEL end][opentelemetry.trace.span.Span.end]"""

        if end_time is None:
            end_time = time.time_ns()

        self._end_timestamp = end_time

        if self.is_recording():
            self.set_status(
                trace_api.status.Status(trace_api.status.StatusCode.OK)
            )

    def __enter__(self) -> Span:
        """See [OTEL __enter__][opentelemetry.trace.span.Span.__enter__]."""

        return self

    def __exit__(
        self,
        exc_type: Optional[BaseException],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        """See [OTEL __exit__][opentelemetry.trace.span.Span.__exit__]."""

        try:
            if exc_val is not None:
                self.record_exception(exception=exc_val)
                raise exc_val
        finally:
            self.end()

    # Rest of these methods are for exporting spans to ReadableSpan. All are not standard OTEL.

    @staticmethod
    def otel_context_of_context(context: SpanContext) -> trace_api.SpanContext:
        return trace_api.SpanContext(
            trace_id=context.trace_id,
            span_id=context.span_id,
            is_remote=False,
        )

    def otel_name(self) -> str:
        return self.name

    def otel_context(self) -> types_api.SpanContext:
        return self.otel_context_of_context(self.context)

    def otel_parent_context(self) -> Optional[types_api.SpanContext]:
        if self.parent is None:
            return None

        return self.otel_context_of_context(self.parent)

    def otel_attributes(self) -> types_api.Attributes:
        return flatten_lensed_attributes(self.attributes)

    def otel_kind(self) -> types_api.SpanKind:
        return trace_api.SpanKind.INTERNAL

    def otel_status(self) -> trace_api.status.Status:
        return trace_api.status.Status(self.status, self.status_description)

    def otel_resource_attributes(self) -> Dict[str, Any]:
        #  TODO(SNOW-1711959)
        return {
            "service.namespace": "trulens",
        }

    def otel_resource(self) -> resources_sdk.Resource:
        return resources_sdk.Resource(
            attributes=self.otel_resource_attributes()
        )

    def otel_events(self) -> List[types_api.Event]:
        return self.events

    def otel_links(self) -> List[types_api.Link]:
        return self.links

    def otel_freeze(self) -> trace_sdk.ReadableSpan:
        """Convert span to an OTEL compatible span for exporting to OTEL collectors."""

        return trace_sdk.ReadableSpan(
            name=self.otel_name(),
            context=self.otel_context(),
            parent=self.otel_parent_context(),
            resource=self.otel_resource(),
            attributes=self.otel_attributes(),
            events=self.otel_events(),
            links=self.otel_links(),
            kind=self.otel_kind(),
            instrumentation_info=None,  # TODO(SNOW-1711959)
            status=self.otel_status(),
            start_time=self.start_timestamp,
            end_time=self.end_timestamp,
            instrumentation_scope=None,  # TODO(SNOW-1711959)
        )


class Tracer(serial_utils.SerialModel, trace_api.Tracer):
    """[OTEL Tracer][opentelemetry.trace.Tracer] requirements."""

    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    _instrumenting_module_name: Optional[str] = pydantic.PrivateAttr(None)
    """Name of the library/module that is instrumenting the code."""

    _instrumenting_library_version: Optional[str] = pydantic.PrivateAttr(None)
    """Version of the library that is instrumenting the code."""

    _attributes: Optional[trace_api.types.Attributes] = pydantic.PrivateAttr(
        None
    )
    """Common attributes to add to all spans."""

    _schema_url: Optional[str] = pydantic.PrivateAttr(None)
    """Use unknown."""

    _tracer_provider: TracerProvider = pydantic.PrivateAttr(None)
    """NON-STANDARD: The TracerProvider that made this tracer."""

    _span_class: Type[Span] = pydantic.PrivateAttr(Span)
    """NON-STANDARD: The default span class to use when creating spans."""

    _span_context_class: Type[SpanContext] = pydantic.PrivateAttr(SpanContext)
    """NON-STANDARD: The default span context class to use when creating spans."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        for k, v in kwargs.items():
            if v is None:
                continue
            # pydantic does not set private attributes in init
            if k.startswith("_") and hasattr(self, k):
                setattr(self, k, v)

    @property
    def context_cvar(
        self,
    ) -> contextvars.ContextVar[context_api.context.Context]:
        return self._tracer_provider.context_cvar

    @property
    def trace_id(self) -> int:
        return self._tracer_provider.trace_id

    def start_span(
        self,
        name: Optional[str] = None,
        *,
        context: Optional[context_api.context.Context] = None,
        kind: trace_api.SpanKind = trace_api.SpanKind.INTERNAL,
        attributes: trace_api.types.Attributes = None,
        links: trace_api._Links = None,
        start_time: Optional[int] = None,
        record_exception: bool = True,
        set_status_on_exception: bool = True,
        cls: Optional[Type[Span]] = None,  # non-standard
    ) -> Span:
        """See [OTEL
        Tracer.start_span][opentelemetry.trace.Tracer.start_span]."""

        if context is None:
            parent_context = self.context_cvar.get()

        else:
            if len(context) != 1:
                raise ValueError("Only one context span is allowed.")
            parent_span_encoding = next(iter(context.values()))

            parent_context = self._span_context_class(
                trace_id=parent_span_encoding.trace_id,
                span_id=parent_span_encoding.span_id,
                _tracer=self,
            )

        new_context = self._span_context_class(
            trace_id=self.trace_id, _tracer=self
        )

        if name is None:
            name = python_utils.class_name(self._span_class)

        if attributes is None:
            attributes = {}

        if self._attributes is not None:
            attributes.update(self._attributes)

        if cls is None:
            cls = self._span_class

        new_span = cls(
            _name=name,
            _context=new_context,
            _parent=parent_context,
            _kind=kind,
            _attributes=attributes,
            _links=links,
            _start_timestamp=start_time,
            _record_exception=record_exception,
            _set_status_on_exception=set_status_on_exception,
            _tracer=self,
        )

        return new_span

    @contextlib.contextmanager
    def start_as_current_span(
        self,
        name: Optional[str] = None,
        context: Optional[trace_api.context.Context] = None,
        kind: trace_api.SpanKind = trace_api.SpanKind.INTERNAL,
        attributes: trace_api.types.Attributes = None,
        links: trace_api._Links = None,
        start_time: Optional[int] = None,
        record_exception: bool = True,
        set_status_on_exception: bool = True,
        end_on_exit: bool = True,
    ):
        """See [OTEL
        Tracer.start_as_current_span][opentelemetry.trace.Tracer.start_as_current_span]."""

        span = self.start_span(
            name=name,
            context=context,
            kind=kind,
            attributes=attributes,
            links=links,
            start_time=start_time,
            record_exception=record_exception,
            set_status_on_exception=set_status_on_exception,
        )

        token = self.context_cvar.set(span.context)

        try:
            yield span

        except BaseException as e:
            if record_exception:
                span.record_exception(e)

        finally:
            self.context_cvar.reset(token)

            if end_on_exit:
                span.end()


class TracerProvider(serial_utils.SerialModel, trace_api.TracerProvider):
    """[OTEL TracerProvider][opentelemetry.trace.TracerProvider]
    requirements."""

    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    _tracer_class: Type[Tracer] = pydantic.PrivateAttr(Tracer)
    """NON-STANDARD: The default tracer class to use when creating tracers."""

    _context_cvar: contextvars.ContextVar[context_api.context.Context] = (
        pydantic.PrivateAttr(
            default_factory=lambda: contextvars.ContextVar(
                "context", default=None
            )
        )
    )

    @property
    def context_cvar(
        self,
    ) -> contextvars.ContextVar[context_api.context.Context]:
        """NON-STANDARD: The context variable to store the current span context."""

        return self._context_cvar

    _trace_id: int = pydantic.PrivateAttr(default_factory=new_trace_id)

    @property
    def trace_id(self) -> int:
        """NON-STANDARD: The current trace id."""

        return self._trace_id

    def get_tracer(
        self,
        instrumenting_module_name: str,
        instrumenting_library_version: Optional[str] = None,
        schema_url: Optional[str] = None,
        attributes: Optional[types_api.Attributes] = None,
    ):
        """See [OTEL
        TracerProvider.get_tracer][opentelemetry.trace.TracerProvider.get_tracer]."""

        tracer = self._tracer_class(
            _instrumenting_module_name=instrumenting_module_name,
            _instrumenting_library_version=instrumenting_library_version,
            _attributes=attributes,
            _schema_url=schema_url,
            _tracer_provider=self,
        )

        return tracer



"""
Utilities for app components provided as part of the trulens package.
Currently organizes all such components as "Other".
"""

from typing import Type

from trulens.core import app
from trulens.core.utils import pyschema as pyschema_utils


class Other(app.Other, app.TrulensComponent):
    pass


# All component types, keep Other as the last one since it always matches.
COMPONENT_VIEWS = [Other]


def constructor_of_class(
    cls_obj: pyschema_utils.Class,
) -> Type[app.TrulensComponent]:
    for view in COMPONENT_VIEWS:
        if view.class_is(cls_obj):
            return view

    raise TypeError(f"Unknown trulens component type with class {cls_obj}")


def component_of_json(json: dict) -> app.TrulensComponent:
    cls = pyschema_utils.Class.of_class_info(json)

    view = constructor_of_class(cls)

    return view(json)



"""Common/shared serializable classes."""

from __future__ import annotations

import datetime
from typing import Optional

import pydantic
from trulens.core.utils import containers as container_utils
from trulens.core.utils import serial as serial_utils

MAX_DILL_SIZE: int = 1024 * 1024  # 1MB
"""Max size in bytes of pickled objects."""


class Cost(serial_utils.SerialModel, pydantic.BaseModel):
    """Costs associated with some call or set of calls."""

    n_requests: int = 0
    """Number of requests."""

    n_successful_requests: int = 0
    """Number of successful requests."""

    n_completion_requests: int = 0
    """Number of completion requests."""

    n_classification_requests: int = 0
    """Number of classification requests."""

    n_classes: int = 0
    """Number of class scores retrieved."""

    n_embedding_requests: int = 0
    """Number of embedding requests."""

    n_embeddings: int = 0
    """Number of embeddings retrieved."""

    n_tokens: int = 0
    """Total tokens processed."""

    n_stream_chunks: int = 0
    """In streaming mode, number of chunks produced."""

    n_prompt_tokens: int = 0
    """Number of prompt tokens supplied."""

    n_completion_tokens: int = 0
    """Number of completion tokens generated."""

    n_cortex_guardrails_tokens: int = 0
    """Number of guardrails tokens generated. i.e. available in Cortex endpoint."""

    cost: float = 0.0
    """Cost in [cost_currency]."""

    cost_currency: str = "USD"

    def __add__(self, other: "Cost") -> "Cost":
        kwargs = {
            k: getattr(self, k) + getattr(other, k)
            if k != "cost_currency"
            and isinstance(getattr(self, k), (int, float))
            else getattr(other, k)
            for k in self.model_fields.keys()
        }
        return Cost(**kwargs)

    def __radd__(self, other: "Cost") -> "Cost":
        # Makes sum work on lists of Cost.

        if other == 0:
            return self

        return self.__add__(other)


class Perf(serial_utils.SerialModel, pydantic.BaseModel):
    """Performance information.

    Presently only the start and end times, and thus latency.
    """

    start_time: datetime.datetime
    """Datetime before the recorded call."""

    end_time: datetime.datetime
    """Datetime after the recorded call."""

    @staticmethod
    def min():
        """Zero-length span with start and end times at the minimum datetime."""

        return Perf(
            start_time=datetime.datetime.min, end_time=datetime.datetime.min
        )

    @staticmethod
    def now(latency: Optional[datetime.timedelta] = None) -> Perf:
        """Create a `Perf` instance starting now and ending now plus latency.

        Args:
            latency: Latency in seconds. If given, end time will be now plus
                latency. Otherwise end time will be a minimal interval plus start_time.
        """

        start_time = datetime.datetime.now()
        if latency is not None:
            end_time = start_time + latency
        else:
            end_time = start_time + datetime.timedelta(microseconds=1)

        return Perf(start_time=start_time, end_time=end_time)

    @property
    def latency(self):
        """Latency in seconds."""
        return self.end_time - self.start_time

    @staticmethod
    def of_ns_timestamps(
        start_ns_timestamp: int, end_ns_timestamp: Optional[int] = None
    ) -> Perf:
        """EXPERIMENTAL: otel-tracing

        Create a `Perf` instance from start and end times in nanoseconds
        since the epoch."""

        return Perf(
            start_time=container_utils.datetime_of_ns_timestamp(
                start_ns_timestamp
            ),
            end_time=container_utils.datetime_of_ns_timestamp(end_ns_timestamp)
            if end_ns_timestamp is not None
            else datetime.datetime.max,
        )

    @property
    def start_ns_timestamp(self) -> int:
        """EXPERIMENTAL: otel-tracing

        Start time in number of nanoseconds since the epoch."""

        return container_utils.ns_timestamp_of_datetime(self.start_time)

    @property
    def end_ns_timestamp(self) -> int:
        """EXPERIMENTAL: otel-tracing

        End time in number of nanoseconds since the epoch."""

        return container_utils.ns_timestamp_of_datetime(self.end_time)


# HACK013: Need these if using __future__.annotations .
Cost.model_rebuild()
Perf.model_rebuild()



from typing import Any, Optional

from trulens.core import experimental as core_experimental
from trulens.core import session as core_session
from trulens.core.utils import python as python_utils
from trulens.core.utils import text as text_utils


class _TruSession(core_session.TruSession):
    def _setup_otel_exporter(
        self, val: Optional[Any]
    ):  # any actually otel_export_sdk.SpanExporter
        if val is None:
            self._experimental_otel_exporter = None
            return

        # from opentelemetry import sdk as otel_sdk
        try:
            from opentelemetry.sdk.trace import export as otel_export_sdk
        except ImportError:
            raise ImportError(
                "OpenTelemetry SDK not found. Please install OpenTelemetry SDK to use OpenTelemetry tracing."
            )

        assert isinstance(
            val, otel_export_sdk.SpanExporter
        ), "otel_exporter must be an OpenTelemetry SpanExporter."

        self._experimental_feature(
            flag=core_experimental.Feature.OTEL_TRACING, value=True, lock=True
        )

        print(
            f"{text_utils.UNICODE_CHECK} OpenTelemetry exporter set: {python_utils.class_name(val.__class__)}"
        )

        self._experimental_otel_exporter = val



"""Implementation of recording that resembles the tracing process in OpenTelemetry.

!!! Note
    Most of the module is (EXPERIMENTAL: otel-tracing) though it includes some existing
    non-experimental classes moved here to resolve some circular import issues.

This module is likely temporary and will be replaced by actual OpenTelemetry sdk
components or implementations that are compatible with its API.
"""

from __future__ import annotations

import contextlib
import contextvars
import functools
import inspect
import logging
import os
import threading as th
from threading import Lock
from types import TracebackType
from typing import (
    Any,
    Callable,
    Dict,
    Hashable,
    Iterable,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)
import uuid

from opentelemetry.semconv.resource import ResourceAttributes
from opentelemetry.trace import span as span_api
from opentelemetry.util import types as types_api
import pydantic
from trulens.core.schema import base as base_schema
from trulens.core.schema import record as record_schema
from trulens.core.schema import types as types_schema
from trulens.core.utils import json as json_utils
from trulens.core.utils import pyschema as pyschema_utils
from trulens.core.utils import python as python_utils
from trulens.core.utils import serial as serial_utils
from trulens.experimental.otel_tracing.core import otel as core_otel
from trulens.experimental.otel_tracing.core._utils import wrap as wrap_utils

logger = logging.getLogger(__name__)

INSTRUMENT: str = "__tru_instrumented"
"""Attribute name to be used to flag instrumented objects/methods/others."""

APPS: str = "__tru_apps"
"""Attribute name for storing apps that expect to be notified of calls."""


class SpanContext(core_otel.SpanContext, Hashable):
    """TruLens additions on top of OTEL SpanContext to add Hashable and
    reference to tracer that made the span."""

    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    def __str__(self):
        return f"{self.trace_id % 0xFF:02x}/{self.span_id % 0xFF:02x}"

    def __repr__(self):
        return str(self)

    def __hash__(self):
        return self.trace_id + self.span_id

    def __eq__(self, other: SpanContextLike):
        if other is None:
            return False

        return self.trace_id == other.trace_id and self.span_id == other.span_id

    @staticmethod
    def of_spancontextlike(span_context: SpanContextLike) -> SpanContext:
        if isinstance(span_context, SpanContext):
            return span_context

        elif isinstance(span_context, core_otel.SpanContext):
            return SpanContext(
                trace_id=span_context.trace_id,
                span_id=span_context.span_id,
                is_remote=span_context.is_remote,
            )
        elif isinstance(span_context, span_api.SpanContext):
            return SpanContext(
                trace_id=span_context.trace_id,
                span_id=span_context.span_id,
                is_remote=span_context.is_remote,
            )
        else:
            raise ValueError(f"Unrecognized span context type: {span_context}")


SpanContextLike = Union[
    SpanContext, core_otel.SpanContext, span_api.SpanContext
]


class Span(core_otel.Span):
    """TruLens additions on top of OTEL spans."""

    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    def __str__(self):
        return (
            f"{type(self).__name__}({self.name}, {self.context}->{self.parent})"
        )

    def __repr__(self):
        return str(self)

    _lensed_attributes: serial_utils.LensedDict[Any] = pydantic.PrivateAttr(
        default_factory=serial_utils.LensedDict
    )

    @property
    def lensed_attributes(self) -> serial_utils.LensedDict[Any]:
        return self._lensed_attributes

    @property
    def parent_span(self) -> Optional[Span]:
        if self.parent is None:
            return None

        if (span := self._tracer.spans.get(self.parent)) is None:
            return None

        return span

    _children_spans: List[Span] = pydantic.PrivateAttr(default_factory=list)

    @property
    def children_spans(self) -> List[Span]:
        return self._children_spans

    error: Optional[Exception] = pydantic.Field(None)
    """Optional error if the observed computation raised an exception."""

    def __init__(self, **kwargs):
        # Convert any contexts to our hashable context class:
        if (context := kwargs["_context"]) is not None:
            kwargs["_context"] = SpanContext.of_spancontextlike(context)
        if (parent := kwargs.get("_parent", None)) is not None:
            kwargs["_parent"] = SpanContext.of_spancontextlike(parent)

        super().__init__(**kwargs)

        if (parent_span := self.parent_span) is not None:
            parent_span.children_spans.append(self)

    def iter_children(
        self, transitive: bool = True, include_phantom: bool = False
    ) -> Iterable[Span]:
        """Iterate over all spans that are children of this span.

        Args:
            transitive: Iterate recursively over children.

            include_phantom: Include phantom spans. If not set, phantom spans
                will not be included but will be iterated over even if
                transitive is false.
        """

        for child_span in self.children_spans:
            if isinstance(child_span, PhantomSpan) and not include_phantom:
                # Note that transitive being false is ignored if phantom is skipped.
                yield from child_span.iter_children(
                    transitive=transitive, include_phantom=include_phantom
                )
            else:
                yield child_span
                if transitive:
                    yield from child_span.iter_children(
                        transitive=transitive,
                        include_phantom=include_phantom,
                    )

    def iter_family(self, include_phantom: bool = False) -> Iterable[Span]:
        """Iterate itself and all children transitively."""

        if (not isinstance(self, PhantomSpan)) or include_phantom:
            yield self

        yield from self.iter_children(
            include_phantom=include_phantom, transitive=True
        )

    def total_cost(self) -> base_schema.Cost:
        """Total costs of this span and all its transitive children."""

        total = base_schema.Cost()

        for span in self.iter_family(include_phantom=True):
            if isinstance(span, WithCost) and span.cost is not None:
                total += span.cost

        return total

    async def __aenter__(self) -> Span:
        return self.__enter__()

    async def __aexit__(
        self,
        exc_type: Optional[BaseException],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        return self.__exit__(exc_type, exc_val, exc_tb)


class PhantomSpan(Span):
    """A span type that indicates that it does not correspond to a
    computation to be recorded but instead is an element of the tracing system.

    It is to be removed from the spans presented to the users.
    """


class LiveSpan(Span):
    """A a span type that indicates that it contains live python objects.

    It is to be converted to a non-live span before being output to the user or
    otherwise.
    """


class PhantomSpanRecordingContext(PhantomSpan):
    """Tracks the context of an app used as a context manager."""

    recording: Optional[Any] = pydantic.Field(None, exclude=True)
    # TODO: app.RecordingContext # circular import issues

    def otel_resource_attributes(self) -> Dict[str, Any]:
        ret = super().otel_resource_attributes()

        ret[ResourceAttributes.SERVICE_NAME] = (
            self.recording.app.app_name if self.recording is not None else None
        )

        return ret

    # override
    def end(self, *args, **kwargs):
        super().end(*args, **kwargs)

        self._finalize_recording()

    # override
    def record_exception(
        self,
        exception: BaseException,
        attributes: types_api.Attributes = None,
        timestamp: int | None = None,
        escaped: bool = False,
    ) -> None:
        super().record_exception(exception, attributes, timestamp, escaped)

        self._finalize_recording()

    def _finalize_recording(self):
        assert self.recording is not None

        app = self.recording.app

        for span in Tracer.find_each_child(
            span=self, span_filter=lambda s: isinstance(s, LiveSpanCall)
        ):
            app._on_new_root_span(recording=self.recording, root_span=span)

        app._on_new_recording_span(recording_span=self)

    def otel_name(self) -> str:
        return "trulens.recording"


class SpanCall(Span):
    """Non-live fields of a function call span."""

    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    call_id: Optional[uuid.UUID] = pydantic.Field(None)
    """Unique identifier for the call."""

    stack: Optional[List[record_schema.RecordAppCallMethod]] = pydantic.Field(
        None
    )
    """Call stack of instrumented methods only."""

    sig: Optional[inspect.Signature] = pydantic.Field(None)
    """Signature of the function."""

    func_name: Optional[str] = None
    """Function name."""

    pid: Optional[int] = None
    """Process id."""

    tid: Optional[int] = None
    """Thread id."""

    def end(self):
        super().end()

        self.set_attribute(ResourceAttributes.PROCESS_PID, self.pid)
        self.set_attribute("thread.id", self.tid)  # TODO: semconv

        self.set_attribute("trulens.call_id", str(self.call_id))
        self.set_attribute("trulens.stack", json_utils.jsonify(self.stack))
        self.set_attribute("trulens.sig", str(self.sig))

    def otel_name(self) -> str:
        return f"trulens.call.{self.func_name}"


class LiveSpanCall(LiveSpan, SpanCall):
    """Track a function call."""

    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    live_obj: Optional[Any] = pydantic.Field(None, exclude=True)
    """Self object if method call."""

    live_cls: Optional[Type] = pydantic.Field(None, exclude=True)
    """Class if method/static/class method call."""

    live_func: Optional[Callable] = pydantic.Field(None, exclude=True)
    """Function object."""

    live_args: Optional[Tuple[Any, ...]] = pydantic.Field(None, exclude=True)
    """Positional arguments to the function call."""

    live_kwargs: Optional[Dict[str, Any]] = pydantic.Field(None, exclude=True)
    """Keyword arguments to the function call."""

    live_bindings: Optional[inspect.BoundArguments] = pydantic.Field(
        None, exclude=True
    )
    """Bound arguments to the function call if can be bound."""

    live_ret: Optional[Any] = pydantic.Field(None, exclude=True)
    """Return value of the function call.

    Exclusive with `error`.
    """

    live_error: Optional[Any] = pydantic.Field(None, exclude=True)
    """Error raised by the function call.

    Exclusive with `ret`.
    """

    def end(self):
        super().end()

        if self.live_cls is not None:
            self.set_attribute(
                "trulens.cls",
                pyschema_utils.Class.of_class(self.live_cls).model_dump(),
            )

        if self.live_func is not None:
            self.set_attribute(
                "trulens.func",
                pyschema_utils.FunctionOrMethod.of_callable(
                    self.live_func
                ).model_dump(),
            )

        if self.live_ret is not None:
            self.set_attribute("trulens.ret", json_utils.jsonify(self.live_ret))

        if self.live_bindings is not None:
            self.set_attribute(
                "trulens.bindings",
                pyschema_utils.Bindings.of_bound_arguments(
                    self.live_bindings, arguments_only=True, skip_self=True
                ).model_dump()["kwargs"],
            )

        if self.live_error is not None:
            self.set_attribute(
                "trulens.error", json_utils.jsonify(self.live_error)
            )


class WithCost(LiveSpan):
    """Mixin to indicate the span has costs tracked."""

    cost: base_schema.Cost = pydantic.Field(default_factory=base_schema.Cost)
    """Cost of the computation spanned."""

    endpoint: Optional[Any] = pydantic.Field(None, exclude=True)
    """Endpoint handling cost extraction for this span/call."""

    def end(self):
        super().end()

        self.set_attribute("trulens.cost", self.cost.model_dump())

    def __init__(self, cost: Optional[base_schema.Cost] = None, **kwargs):
        if cost is None:
            cost = base_schema.Cost()

        super().__init__(cost=cost, **kwargs)


class LiveSpanCallWithCost(LiveSpanCall, WithCost):
    pass


class Tracer(core_otel.Tracer):
    """TruLens additions on top of [OTEL Tracer][opentelemetry.trace.Tracer]."""

    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    # Overrides core_otel.Tracer._span_class
    _span_class: Type[Span] = pydantic.PrivateAttr(Span)

    # Overrides core_otel.Tracer._span_context_class
    _span_context_class: Type[SpanContext] = pydantic.PrivateAttr(SpanContext)

    @property
    def spans(self) -> Dict[SpanContext, Span]:
        return self._tracer_provider.spans

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __str__(self):
        return f"{type(self).__name__} {self.instrumenting_module_name} {self.instrumenting_library_version}"

    def __repr__(self):
        return str(self)

    def start_span(self, *args, **kwargs):
        new_span = super().start_span(*args, **kwargs)

        self.spans[new_span.context] = new_span

        return new_span

    @staticmethod
    def _fill_stacks(
        span: Span,
        get_method_path: Callable,
        stack: List[record_schema.RecordAppCallMethod] = [],
    ):
        if isinstance(span, LiveSpanCall):
            path = get_method_path(obj=span.live_obj, func=span.live_func)

            frame_ident = record_schema.RecordAppCallMethod(
                path=path
                if path is not None
                else serial_utils.Lens().static,  # placeholder path for functions
                method=pyschema_utils.Method.of_method(
                    span.live_func, obj=span.live_obj, cls=span.live_cls
                ),
            )

            stack = stack + [frame_ident]
            span.stack = stack

        for subspan in span.iter_children(transitive=False):
            Tracer._fill_stacks(
                subspan, stack=stack, get_method_path=get_method_path
            )

    def _call_of_spancall(
        self, span: LiveSpanCall
    ) -> record_schema.RecordAppCall:
        """Convert a SpanCall to a RecordAppCall."""

        args = (
            dict(span.live_bindings.arguments)
            if span.live_bindings is not None
            else None
        )
        if args is not None:
            if "self" in args:
                del args["self"]  # remove self

        assert span.start_timestamp is not None
        assert span.end_timestamp is not None

        return record_schema.RecordAppCall(
            call_id=str(span.call_id),
            stack=span.stack,
            args=args,
            rets=json_utils.jsonify(span.live_ret),
            error=str(span.live_error),
            perf=base_schema.Perf.of_ns_timestamps(
                start_ns_timestamp=span.start_timestamp,
                end_ns_timestamp=span.end_timestamp,
            ),
            pid=span.pid,
            tid=span.tid,
        )

    def record_of_root_span(
        self, recording: Any, root_span: LiveSpanCall
    ) -> record_schema.Record:
        """Convert a root span to a record.

        This span has to be a call span so we can extract things like main input and output.
        """

        assert isinstance(root_span, LiveSpanCall), type(root_span)

        app = recording.app

        self._fill_stacks(root_span, get_method_path=app.get_method_path)

        root_perf = (
            base_schema.Perf.of_ns_timestamps(
                start_ns_timestamp=root_span.start_timestamp,
                end_ns_timestamp=root_span.end_timestamp,
            )
            if root_span.end_timestamp is not None
            else None
        )

        if isinstance(root_span, WithCost):
            root_cost = root_span.total_cost()
        else:
            root_cost = base_schema.Cost()

        calls = []
        if isinstance(root_span, LiveSpanCall):
            calls.append(self._call_of_spancall(root_span))

        spans = [root_span]

        for span in root_span.iter_children(include_phantom=True):
            if isinstance(span, LiveSpanCall):
                calls.append(self._call_of_spancall(span))

            spans.append(span)

        bindings = root_span.live_bindings
        main_error = root_span.live_error

        if bindings is not None:
            main_input = app.main_input(
                func=root_span.live_func,
                sig=root_span.sig,
                bindings=root_span.live_bindings,
            )
            if main_error is None:
                main_output = app.main_output(
                    func=root_span.live_func,
                    sig=root_span.sig,
                    bindings=root_span.live_bindings,
                    ret=root_span.live_ret,
                )
            else:
                main_output = None
        else:
            main_input = None
            main_output = None

        record = record_schema.Record(
            record_id="placeholder",
            app_id=app.app_id,
            main_input=json_utils.jsonify(main_input),
            main_output=json_utils.jsonify(main_output),
            main_error=json_utils.jsonify(main_error),
            calls=calls,
            perf=root_perf,
            cost=root_cost,
            experimental_otel_spans=spans,
        )

        # record_id determinism
        record.record_id = json_utils.obj_id_of_obj(record, prefix="record")

        return record

    @staticmethod
    def find_each_child(span: Span, span_filter: Callable) -> Iterable[Span]:
        """For each family rooted at each child of this span, find the top-most
        span that satisfies the filter."""

        for child_span in span.children_spans:
            if span_filter(child_span):
                yield child_span
            else:
                yield from Tracer.find_each_child(child_span, span_filter)

    def records_of_recording(
        self, recording: PhantomSpanRecordingContext
    ) -> Iterable[record_schema.Record]:
        """Convert a recording based on spans to a list of records."""

        for root_span in Tracer.find_each_child(
            span=recording, span_filter=lambda s: isinstance(s, LiveSpanCall)
        ):
            assert isinstance(root_span, LiveSpanCall)
            yield self.record_of_root_span(
                recording=recording, root_span=root_span
            )

    @contextlib.contextmanager
    def _span(self, cls, **kwargs):
        with self.start_span(cls=cls, **kwargs) as span:
            token = self.context_cvar.set(span.context)

            try:
                yield span

            finally:
                self.context_cvar.reset(token)

    @contextlib.asynccontextmanager
    async def _aspan(self, cls, **kwargs):
        async with self.start_span(cls=cls, **kwargs) as span:
            token = self.context_cvar.set(span.context)

            try:
                yield span

            finally:
                self.context_cvar.reset(token)

    # context manager
    def recording(self):
        return self._span(
            name="trulens.recording", cls=PhantomSpanRecordingContext
        )

    # context manager
    def method(self, method_name: str):
        return self._span(name="trulens.call." + method_name, cls=LiveSpanCall)

    # context manager
    def cost(self, method_name: str, cost: Optional[base_schema.Cost] = None):
        return self._span(
            name="trulens.call." + method_name,
            cls=LiveSpanCallWithCost,
            cost=cost,
        )

    # context manager
    def phantom(self):
        return self._span(name="trulens.phantom", cls=PhantomSpan)

    # context manager
    async def arecording(self):
        return self._aspan(
            name="trulens.recording", cls=PhantomSpanRecordingContext
        )

    # context manager
    async def amethod(self, method_name: str):
        return self._aspan(name="trulens.call." + method_name, cls=LiveSpanCall)

    # context manager
    async def acost(
        self, method_name: str, cost: Optional[base_schema.Cost] = None
    ):
        return self._aspan(
            name="trulens.call." + method_name,
            cls=LiveSpanCallWithCost,
            cost=cost,
        )

    # context manager
    async def aphantom(self):
        return self._aspan(name="trulens.phantom", cls=PhantomSpan)


class NullTracer(Tracer):
    """Tracer that does not save the spans it makes."""

    @contextlib.contextmanager
    def _span(self, cls, **kwargs):
        # TODO: this adds span to global list, don't do this
        span = self.start_span(cls=cls, **kwargs)

        token = self.context.set(span.context)

        error = None

        try:
            yield span
        except BaseException as e:
            # ignore exception since spans are also ignored/not recorded
            error = e
        finally:
            self.context.reset(token)
            if error is not None:
                raise error

    @contextlib.asynccontextmanager
    async def _aspan(self, cls, **kwargs):
        span = self.start_span(cls=cls, **kwargs)
        token = self.context.set(span.context)

        error = None

        try:
            yield span
        except BaseException as e:
            # ignore exception since spans are also ignored/not recorded
            error = e
        finally:
            self.context.reset(token)
            if error is not None:
                raise error

    @property
    def spans(self):
        return []


class TracerProvider(
    core_otel.TracerProvider, metaclass=python_utils.PydanticSingletonMeta
):
    """TruLens additions on top of [OTEL TracerProvider][opentelemetry.trace.TracerProvider]."""

    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    _trace_id: int = pydantic.PrivateAttr(
        default_factory=core_otel.new_trace_id
    )

    def __str__(self):
        # Pydantic will not print anything useful otherwise.
        return f"{self.__module__}.{type(self).__name__}()"

    @property
    def trace_id(self):
        return self._trace_id

    # Overrides core_otel.TracerProvider._tracer_class
    _tracer_class: Type[Tracer] = pydantic.PrivateAttr(default=Tracer)

    _tracers: Dict[str, Tracer] = pydantic.PrivateAttr(default_factory=dict)

    _spans: Dict[SpanContext, Span] = pydantic.PrivateAttr(default_factory=dict)

    @property
    def spans(self) -> Dict[SpanContext, Span]:
        return self._spans

    def get_tracer(
        self,
        instrumenting_module_name: str,
        instrumenting_library_version: Optional[str] = None,
        schema_url: Optional[str] = None,
        attributes: Optional[types_api.Attributes] = None,
    ):
        if instrumenting_module_name in self._tracers:
            return self._tracers[instrumenting_module_name]

        tracer = super().get_tracer(
            instrumenting_module_name=instrumenting_module_name,
            instrumenting_library_version=instrumenting_library_version,
            attributes=attributes,
            schema_url=schema_url,
        )

        self._tracers[instrumenting_module_name] = tracer

        return tracer


tracer_provider = TracerProvider()
"""Global tracer provider.
All trulens tracers are made by this provider even if a different one is
configured for OTEL.
"""


@functools.cache
def trulens_tracer():
    from trulens.core import __version__

    return tracer_provider.get_tracer(
        instrumenting_module_name="trulens.experimental.otel_tracing.core.trace",
        instrumenting_library_version=__version__,
    )


T = TypeVar("T")


class TracingCallbacks(wrap_utils.CallableCallbacks[T]):
    """Extension of CallableCallbacks that adds tracing to the wrapped callable
    as implemented using tracer and spans."""

    @classmethod
    def on_callable_wrapped(
        cls, wrapper: Callable[..., Any], **kwargs: Dict[str, Any]
    ):
        return super().on_callable_wrapped(wrapper=wrapper, **kwargs)

    def __init__(
        self,
        func_name: str,
        span_type: Type[LiveSpanCall] = LiveSpanCall,
        **kwargs: Dict[str, Any],
    ):
        super().__init__(**kwargs)

        self.func_name: str = func_name

        self.obj: Optional[object] = None
        self.obj_cls: Optional[Type] = None
        self.obj_id: Optional[int] = None

        if not issubclass(span_type, LiveSpanCall):
            raise ValueError("span_type must be a subclass of LiveSpanCall.")

        self.span_context = trulens_tracer()._span(
            span_type, name="trulens.call." + func_name
        )
        self.span = self.span_context.__enter__()

    def on_callable_call(
        self, bindings: inspect.BoundArguments, **kwargs: Dict[str, Any]
    ) -> inspect.BoundArguments:
        temp = super().on_callable_call(bindings=bindings, **kwargs)

        if "self" in bindings.arguments:
            self.obj = bindings.arguments["self"]
            self.obj_cls = type(self.obj)
            self.obj_id = id(self.obj)

        span = self.span
        span.pid = os.getpid()
        span.tid = th.get_native_id()

        return temp

    def on_callable_end(self):
        span = self.span

        # SpanCall attributes
        span.call_id = self.call_id
        span.func_name = self.func_name
        span.sig = self.sig

        # LiveSpanCall attributes
        span.live_obj = self.obj
        span.live_cls = self.obj_cls
        span.live_func = self.func
        span.live_args = self.call_args
        span.live_kwargs = self.call_kwargs
        span.live_bindings = self.bindings
        span.live_ret = self.ret
        span.live_error = self.error

        if self.error is not None:
            self.span_context.__exit__(
                type(self.error), self.error, self.error.__traceback__
            )
        else:
            self.span_context.__exit__(None, None, None)


class _RecordingContext:
    """Manager of the creation of records from record calls.

    An instance of this class is produced when using an
    [App][trulens_eval.app.App] as a context mananger, i.e.:
    Example:
        ```python
        app = ...  # your app
        truapp: TruChain = TruChain(app, ...) # recorder for LangChain apps
        with truapp as recorder:
            app.invoke(...) # use your app
        recorder: RecordingContext
        ```

    Each instance of this class produces a record for every "root" instrumented
    method called. Root method here means the first instrumented method in a
    call stack. Note that there may be more than one of these contexts in play
    at the same time due to:
    - More than one wrapper of the same app.
    - More than one context manager ("with" statement) surrounding calls to the
      same app.
    - Calls to "with_record" on methods that themselves contain recording.
    - Calls to apps that use trulens internally to track records in any of the
      supported ways.
    - Combinations of the above.
    """

    def __init__(
        self,
        app: _WithInstrumentCallbacks,
        record_metadata: serial_utils.JSON = None,
        tracer: Optional[Tracer] = None,
        span: Optional[PhantomSpanRecordingContext] = None,
        span_ctx: Optional[SpanContext] = None,
    ):
        self.calls: Dict[types_schema.CallID, record_schema.RecordAppCall] = {}
        """A record (in terms of its RecordAppCall) in process of being created.

        Storing as a map as we want to override calls with the same id which may
        happen due to methods producing awaitables or generators. These result
        in calls before the awaitables are awaited and then get updated after
        the result is ready.
        """
        # TODEP: To deprecated after migration to span-based tracing.

        self.records: List[record_schema.Record] = []
        """Completed records."""

        self.lock: Lock = Lock()
        """Lock blocking access to `records` when adding calls or
        finishing a record."""

        self.token: Optional[contextvars.Token] = None
        """Token for context management."""

        self.app: _WithInstrumentCallbacks = app
        """App for which we are recording."""

        self.record_metadata = record_metadata
        """Metadata to attach to all records produced in this context."""

        self.tracer: Optional[Tracer] = tracer
        """EXPERIMENTAL(otel-tracing): OTEL-like tracer for recording.
        """

        self.span: Optional[PhantomSpanRecordingContext] = span
        """EXPERIMENTAL(otel-tracing): Span that represents a recording context
        (the with block)."""

        self.span_ctx = span_ctx
        """EXPERIMENTAL(otel-tracing): The context manager for the above span.
        """

    @property
    def spans(self) -> Dict[SpanContext, Span]:
        """EXPERIMENTAL(otel-tracing): Get the spans of the tracer in this context."""

        if self.tracer is None:
            return {}

        return self.tracer.spans

    def __iter__(self):
        return iter(self.records)

    def get(self) -> record_schema.Record:
        """Get the single record only if there was exactly one or throw
        an error otherwise."""

        if len(self.records) == 0:
            raise RuntimeError("Recording context did not record any records.")

        if len(self.records) > 1:
            raise RuntimeError(
                "Recording context recorded more than 1 record. "
                "You can get them with ctx.records, ctx[i], or `for r in ctx: ...`."
            )

        return self.records[0]

    def __getitem__(self, idx: int) -> record_schema.Record:
        return self.records[idx]

    def __len__(self):
        return len(self.records)

    def __hash__(self) -> int:
        # The same app can have multiple recording contexts.
        return hash(id(self.app)) + hash(id(self.records))

    def __eq__(self, other):
        return hash(self) == hash(other)

    def add_call(self, call: record_schema.RecordAppCall):
        """Add the given call to the currently tracked call list."""
        # TODEP: To deprecated after migration to span-based tracing.

        with self.lock:
            # NOTE: This might override existing call record which happens when
            # processing calls with awaitable or generator results.
            self.calls[call.call_id] = call

    def finish_record(
        self,
        calls_to_record: Callable[
            [
                List[record_schema.RecordAppCall],
                types_schema.Metadata,
                Optional[record_schema.Record],
            ],
            record_schema.Record,
        ],
        existing_record: Optional[record_schema.Record] = None,
    ):
        """Run the given function to build a record from the tracked calls and any
        pre-specified metadata."""
        # TODEP: To deprecated after migration to span-based tracing.

        with self.lock:
            record = calls_to_record(
                list(self.calls.values()), self.record_metadata, existing_record
            )
            self.calls = {}

            if existing_record is None:
                # If existing record was given, we assume it was already
                # inserted into this list.
                self.records.append(record)

        return record


class _WithInstrumentCallbacks:
    """Abstract definition of callbacks invoked by Instrument during
    instrumentation or when instrumented methods are called.

    Needs to be mixed into [App][trulens_eval.app.App].
    """

    # Called during instrumentation.
    def on_method_instrumented(
        self, obj: object, func: Callable, path: serial_utils.Lens
    ):
        """Callback to be called by instrumentation system for every function
        requested to be instrumented.

        Given are the object of the class in which `func` belongs
        (i.e. the "self" for that function), the `func` itsels, and the `path`
        of the owner object in the app hierarchy.

        Args:
            obj: The object of the class in which `func` belongs (i.e. the
                "self" for that method).

            func: The function that was instrumented. Expects the unbound
                version (self not yet bound).

            path: The path of the owner object in the app hierarchy.
        """

        raise NotImplementedError

    # Called during invocation.
    def get_method_path(self, obj: object, func: Callable) -> serial_utils.Lens:
        """Get the path of the instrumented function `func`, a member of the class
        of `obj` relative to this app.

        Args:
            obj: The object of the class in which `func` belongs (i.e. the
                "self" for that method).

            func: The function that was instrumented. Expects the unbound
                version (self not yet bound).
        """

        raise NotImplementedError

    # WithInstrumentCallbacks requirement
    def get_methods_for_func(
        self, func: Callable
    ) -> Iterable[Tuple[int, Callable, serial_utils.Lens]]:
        """EXPERIMENTAL(otel-tracing): Get the methods (rather the inner
        functions) matching the given `func` and the path of each.

        Args:
            func: The function to match.
        """

        raise NotImplementedError

    # Called after recording of an invocation.
    def _on_new_root_span(
        self,
        ctx: _RecordingContext,
        root_span: LiveSpanCall,
    ) -> record_schema.Record:
        """EXPERIMENTAL(otel-tracing): Called by instrumented methods if they
        are root calls (first instrumented methods in a call stack).

        Args:
            ctx: The context of the recording.

            root_span: The root span that was recorded.
        """
        # EXPERIMENTAL: otel-tracing

        raise NotImplementedError


class AppTracingCallbacks(TracingCallbacks[T]):
    """Extension to TracingCallbacks that keep track of apps that are
    instrumenting their constituent calls."""

    @classmethod
    def on_callable_wrapped(
        cls,
        wrapper: Callable[..., Any],
        app: _WithInstrumentCallbacks,
        **kwargs: Dict[str, Any],
    ):
        if not python_utils.safe_hasattr(wrapper, APPS):
            apps = set()
            setattr(wrapper, APPS, apps)
        else:
            apps = python_utils.safe_getattr(wrapper, APPS)

        apps.add(app)

        return super().on_callable_wrapped(wrapper, **kwargs)

    def __init__(
        self,
        app: _WithInstrumentCallbacks,
        span_type: Type[Span] = LiveSpanCall,
        **kwargs: Dict[str, Any],
    ):
        super().__init__(span_type=span_type, **kwargs)

        self.app = app
        self.apps = python_utils.safe_getattr(self.wrapper, APPS)
